#include "coupledtimestepper.hh"

template <class Factory, class Updaters, class ErrorNorm>
class AdaptiveTimeStepper {
  using Vector = typename Factory::Vector;
  using ConvexProblem = typename Factory::ConvexProblem;
  using Nonlinearity = typename ConvexProblem::NonlinearityType;

  using MyCoupledTimeStepper = CoupledTimeStepper<Factory, Updaters, ErrorNorm>;

public:
  AdaptiveTimeStepper(Factory &factory, Dune::ParameterTree const &parset,
                      std::shared_ptr<Nonlinearity> globalFriction,
                      Updaters &current, double relativeTime,
                      double relativeTau,
                      std::function<void(double, Vector &)> externalForces,
                      ErrorNorm const &errorNorm,
                      std::function<bool(Updaters &, Updaters &)> mustRefine)
      : relativeTime_(relativeTime),
        relativeTau_(relativeTau),
        finalTime_(parset.get<double>("problem.finalTime")),
        factory_(factory),
        parset_(parset),
        globalFriction_(globalFriction),
        current_(current),
        R1_(current_.clone()),
        externalForces_(externalForces),
        mustRefine_(mustRefine),
        iterationWriter_("iterations", std::fstream::out),
        errorNorm_(errorNorm) {
    MyCoupledTimeStepper coupledTimeStepper(finalTime_, factory_, parset_,
                                            globalFriction_, R1_, errorNorm,
                                            externalForces_);
    stepAndReport("R1", coupledTimeStepper, relativeTime_, relativeTau_);
    iterationWriter_ << std::endl;
  }

  bool reachedEnd() { return relativeTime_ >= 1.0; }

  bool coarsen() {
    bool didCoarsen = false;

    while (relativeTime_ + relativeTau_ < 1.0) {
      R2_ = R1_.clone();
      {
        MyCoupledTimeStepper coupledTimeStepper(finalTime_, factory_, parset_,
                                                globalFriction_, R2_,
                                                errorNorm_, externalForces_);
        stepAndReport("R2", coupledTimeStepper, relativeTime_ + relativeTau_,
                      relativeTau_);
      }

      Updaters C = current_.clone();
      {
        MyCoupledTimeStepper coupledTimeStepper(finalTime_, factory_, parset_,
                                                globalFriction_, C, errorNorm_,
                                                externalForces_);

        stepAndReport("C", coupledTimeStepper, relativeTime_,
                      2.0 * relativeTau_);
      }

      if (!mustRefine_(C, R2_)) {
        R2_ = { nullptr, nullptr };
        R1_ = C;
        relativeTau_ *= 2.0;
        didCoarsen = true;
      } else {
        break;
      }
    }
    return didCoarsen;
  }

  void refine() {
    while (true) {
      Updaters F2 = current_.clone();
      Updaters F1;
      {
        MyCoupledTimeStepper coupledTimeStepper(finalTime_, factory_, parset_,
                                                globalFriction_, F2, errorNorm_,
                                                externalForces_);
        stepAndReport("F1", coupledTimeStepper, relativeTime_,
                      relativeTau_ / 2.0);

        F1 = F2.clone();
        stepAndReport("F2", coupledTimeStepper,
                      relativeTime_ + relativeTau_ / 2.0, relativeTau_ / 2.0);
      }

      if (!mustRefine_(R1_, F2)) {
        break;
      } else {
        R1_ = F1;
        R2_ = F2;
        relativeTau_ /= 2.0;
      }
    }
  }

  void advance() {
    if (!coarsen())
      refine();

    iterationWriter_ << std::endl;

    current_ = R1_;
    R1_ = R2_;
    relativeTime_ += relativeTau_;
  }

  double relativeTime_;
  double relativeTau_;

private:
  void stepAndReport(std::string type, MyCoupledTimeStepper &stepper,
                     double rTime, double rTau) {
    iterationWriter_ << type << " " << stepper.step(rTime, rTau) << " "
                     << std::flush;
  }

  double finalTime_;
  Factory &factory_;
  Dune::ParameterTree const &parset_;
  std::shared_ptr<Nonlinearity> globalFriction_;
  Updaters &current_;
  Updaters R1_;
  Updaters R2_;
  std::function<void(double, Vector &)> externalForces_;
  std::function<bool(Updaters &, Updaters &)> mustRefine_;
  std::fstream iterationWriter_;
  ErrorNorm const &errorNorm_;
};
